/**
 * 
 */
package sim.math;

import java.util.Arrays;

import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;

/**
 * La classe <b>SMath</b> contient des mthodes de calcul qui sont complmentaire  la classe java.lang.Math. 
 * Elle perment entre autre d'valuer les racines relles d'un polynme de degr 1, 2, 3 et 4.
 * 
 * @author Simon Vzina
 * @since 2015-02-19
 * @version 2017-09-20 (labo - La loi de Coulomb v1.1)
 */
public final class SMath {

	/**
	 * <p>
	 * La constante <b>EPSILON</b> reprsentante un nombre trs petit, mais non nul. Ce chiffre peut tre utilis 
	 * pour comparer une valeur de type double avec le chiffre zro. Puisqu'un double gale  zro
	 * est difficile  obtenir numriquement aprs un calcul (sauf si l'on multiplie par zro), il est prfrable de 
	 * comparer un "pseudo zro" avec cette constante.
	 * </p>
	 * <p>
	 * Avec une valeur de EPSILON = 1e-10, cette valeur permet de comparer adquatement des nombres autour de '1' avec suffisamment de chiffres significatifs.
	 * </p>
	 */
	public static double EPSILON = 1e-10;           
	
	/**
   * La constante <b>NEGATIVE_EPSILON</b> reprsentante un nombre trs petit, mais non nul qui est <b>negatif</b>. Ce chiffre peut tre utilis 
   * pour comparer une valeur arbiraire de type double avec le chiffre zro, mais qui sera ngatif. Puisqu'un double gale  zro
   * est difficile  obtenir numriquement aprs un calcul (sauf si l'on multiplie par zro), il est prfrable de 
   * comparer un "pseudo zro" avec cette constante.
   */
	public static double NEGATIVE_EPSILON = -1.0*EPSILON;  
	
	/**
	 * La constante <b>ONE_PLUS_EPSILON</b> reprsente une constante gale  <b>1 + EPSILON</b> ce qui correspond  un nombre lgrement suprieur  1.
	 */
	public static double ONE_PLUS_EPSILON = 1 + EPSILON;
	
	/**
	 * La constante <b>ONE_MINUS_EPSILON</b> reprsente une constant gale  <b>1 - EPSILON</b> ce qui correspond  un nombre lgrement infrieur  1.
	 */
	public static double ONE_MINUS_EPSILON = 1 - EPSILON;
	
	/**
   * La constante <b>ONE_PLUS_1000EPSILON</b> reprsente une constante gale  <b>1 + 1000*EPSILON</b> ce qui correspond  un nombre lgrement suprieur  1.
   */
  public static double ONE_PLUS_1000EPSILON = 1 + 1000*EPSILON;
  
  /**
   * La constante <b>ONE_MINUS_EPSILON</b> reprsente une constant gale  <b>1 - 1000*EPSILON</b> ce qui correspond  un nombre lgrement infrieur  1.
   */
  public static double ONE_MINUS_1000EPSILON = 1 - 1000*EPSILON;
  
	/**
	 * La constante <b>INFINITY</b> reprsente un nombre positif gale  l'infini. Cette valeur est obtenue 
	 * partir de la classe java.lang.Double.
	 * @see java.lang.Double
	 */
	public static double INFINITY = Double.POSITIVE_INFINITY;
	
	/**
   * Mthode pour dterminer si deux nombres de type double sont <b>relativement gaux</b>. 
   * En utilisant une approche de calcul de diffrence, on vrifie si
   * <ul>a - b < EPSILON*ref</ul>  
   * afin de <b>valid l'galit</b> entre a et b (a == b). EPSILON est un seuil de prcision 
   * et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b). 
   * <p>Cependant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.</p>
   * 
   * @param a - Le 1ier nombre  comparer.
   * @param b - Le 2ime nombre  comparer.
   * @return <b>true</b> si les deux nombres sont <b>relativement gaux</b> et <b>false</b> sinon.
   */
	public static boolean nearlyEquals(double a, double b)
	{
	  return nearlyEquals(a, b, EPSILON);
	}
	
	/**
	 * Mthode pour dterminer si deux nombres de type double sont <b>relativement gaux</b>. 
	 * En utilisant une approche de calcul de diffrence, on vrifie si
	 * <ul>a - b < EPSILON*ref</ul>  
	 * afin de <b>valid l'galit</b> entre a et b (a == b). EPSILON est un seuil de prcision 
	 * et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b).
	 * <p>Cenpendant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.</p>
	 * 
	 * @param a - Le 1ier nombre  comparer.
	 * @param b - Le 2ime nombre  comparer.
	 * @param epsilon - La prcision acceptable.
	 * @return <b>true</b> si les deux nombres sont <b>relativement gaux</b> et <b>false</b> sinon.
	 */
	public static boolean nearlyEquals(double a, double b, double epsilon)
	{
	  double absA = Math.abs(a);
    double absB = Math.abs(b);
    double diff = Math.abs(a - b);
    
    // Vrification du cas particulier : 0 = 0 et infiny = infiny (mais pas certain ...) 
    if(a == b)                            
      return true;
     
    // Cas des petites chiffres : Vrifier si les deux chiffres sont trs prs l'un de l'autre
    if(diff < epsilon)  
      return true;
      
    // Cas gnral
    double positive_max;
        
    if(absA > absB)
      positive_max = absA;
    else
      positive_max = absB;
      
    if(diff < positive_max*epsilon)
      return true;
    else
      return false;
	}
	
	/**
	 * Mthpde pour dterminer si une valeur est relativement prs de zro.
	 * Cette mthode est ncessaire, car une opration mathmatique menant au chiffre 0 peut tre 0.0 et -0.0 ce qui n'est pas gal selon JAVA.
	 * 
	 * @param value La valeur  comparer avec 0.
	 * @return <b>true</b> si la valeur est <b>relativement gal</b>  zro et <b>false</b> sinon.
	 */
	public static boolean nearlyZero(double value)
	{
	  return nearlyEquals(value, 0.0);
	}
	
	/**
	 * Mthode pour dterminer si deux tableau de nombres de type double sont <b>relativement gaux</b>. 
   * En utilisant une approche de calcul de diffrence, on vrifie si pour chaque lment du tableau
   * <ul>a - b < EPSILON*ref</ul>  
   * afin de <b>valid l'galit</b> entre a et b (a == b). EPSILON est un seuil de prcision 
   * et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b).
   * <p>Cenpendant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.</p>
   * 
	 * @param tab1 Le premier tableau  comparer.
	 * @param tab2 Le deuxime tableau  comparer.
	 * @param epsilon La prcision acceptable.
	 * @return <b>true</b> si les deux tableaux sont <b>relativement gaux</b> et <b>false</b> sinon.
	 */
	public static boolean nearlyEquals(double[] tab1, double[] tab2, double epsilon)
	{
	  // Vrifier que les deux tableaux ont la mme taille.
	  if(tab1.length != tab2.length)
	    return false;
	  
	  // Vrifier que l'ensemble des donne indexe par indexe sont "nearlyEquals".
	  for(int i = 0; i < tab1.length; i++)
	    if(!nearlyEquals(tab1[i], tab2[i], epsilon))
	      return false;
	  
	  // Puisque l'ensemble des valeurs indexe par indexe sont "nearlyEquals", le tableau doit l'tre. 
	  return true;
	}
	
	/**
	 * Mthode permettant d'valuer la racine relle d'un polynme de degr '1' de la forme
	 * <ul>Ax + B = 0.</ul>
	 * <p>Un polynme de degr '1' possde au maximum <b>une</b> racine relle.</p>
	 *  
	 * @param A - Le coefficient devant le terme de puissance '1' (x).
	 * @param B - Le coefficient devant le terme de puissance '0' (1).
	 * @return La racine relle d'un polynme de degr '1' (s'il y en a).
	 */
	public static double[] linearRealRoot(double A, double B)
	{
	  // Vrifier si le polynme n'est pas d'un degr infrieur.
	  // Dans ce cas, la fonction serait constante, donc avec aucune racine possible.
	  if(A == 0.0)
	    return new double[0];
	 
	
    throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
   
	}
	
	/**
   * Mthode permettant d'valuer les racines relles d'un polynme de degr '2' de la forme
   * <ul>Ax^2 + Bx + C = 0.</ul>
   * <p>Un polynme de degr '2' possde au maximum <b>deux</b> racines relles.</p>
   *  
   * @param A - Le coefficient devant le terme de puissance '2' (x^2).
   * @param B - Le coefficient devant le terme de puissance '1' (x).
   * @param C - Le coefficient devant le terme de puissance '0' (1).
   * @return Les racines relles de d'un polynme de degr '2' (s'il y en a). Les solutions retournes dans un tableau sont <b>tries en ordre croissant</b>. 
   */
	public static double[] quadricRealRoot(double A, double B, double C)
	{
	  // Vrifier si le polynme n'est pas d'un degr infrieur
	  if(A == 0.0)
	    return linearRealRoot(B,C);
	  
	  throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
	  
	}
	
	/**
   * Mthode permettant d'valuer les racines relles d'un polynme de degr '3' de la forme
   * <ul>Ax^3 + Bx^2 + Cx + D = 0.</ul>
   * <p>Un polynme de degr '3' possde au maximum <b>trois</b> racines relles.</p>
   *  
   * @param A - Le coefficient devant le terme de puissance '3' (x^3).
   * @param B - Le coefficient devant le terme de puissance '2' (x^2).
   * @param C - Le coefficient devant le terme de puissance '1' (x).
   * @param D - Le coefficient devant le terme de puissance '0' (1).
   * @return Les racines relles de d'un polynme de degr '3' (s'il y en a). Les solutions retournes dans un tableau sont <b>tries en ordre croissant</b>. 
   */
  public static double[] cubicRealRoot(double A, double B, double C, double D)
  {
    // Vrifier si le polynme n'est pas d'un degr infrieur
    if(A == 0.0)
      return quadricRealRoot(B,C,D);
    
    throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
    
  }
  
	/**
   * Mthode permettant d'valuer les racines relles d'un polynme de degr '4' de la forme
   * <ul>Ax^4 + Bx^3 + Cx^2 + Dx + E = 0.</ul>
   * <p>Un polynme de degr '4' possde au maximum <b>quatre</b> racines relles.</p>
   *  
   * @param A - Le coefficient devant le terme de puissance '4' (x^4).
   * @param B - Le coefficient devant le terme de puissance '3' (x^3).
   * @param C - Le coefficient devant le terme de puissance '2' (x^2).
   * @param D - Le coefficient devant le terme de puissance '1' (x).
   * @param E - Le coefficient devant le terme de puissance '0' (1).
   * @return Les racines relles de d'un polynme de degr '4' (s'il y en a). Les solutions retournes dans un tableau sont <b>tries en ordre croissant</b>. 
   */
	public static double[] quarticRealRoot(double A, double B, double C, double D, double E)
	{
	  // Vrifier si le polynme n'est pas d'un degr infrieur.
	  if(A == 0.0)
	    return cubicRealRoot(B,C,D,E);
	  
	 throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
    
	}
	
	/**
   * Mthode qui effectue le calcul inverse de l'interpolation linaire d'une valeur numrique.
   * Cela sigifie que l'on cherche la valeur du paramtre d'interpolation t  partir d'une valeur interpole
   * ainsi que des deux valeurs extrmes.
   * 
   * @param v La valeur interpole dont la valeur de t doit tre calcule.
   * @param v0 La valeur de rfrence pondre par 1 - t.
   * @param v1 La valeur de rfrence pondre par le facteur t.
   * @return La facteur t d'interpolation linaire.
   */
  public static double reverseLinearInterpolation(double v, double v0, double v1) 
  {
    //  partir de la relation v = (1-t)*v0 + t*v1 , on doit isoler t.
    return (v - v0)/(v1 - v0);
  }
  
  /**
   * Mthode qui effectue le calcul de l'interpolation linaire d'une valeur numrique.
   * 
   * @param v0 La valeur de rfrence pondre par 1 - t.
   * @param v1 La valeur de rfrence pondre par le facteur t.
   * @param t Le paramtre de pondration.
   * @return La valeur interpole.
   * @throws SRuntimeException Si la contrainte sur t n'est pas respecte (0 <= t <= 1).
   */
  public static double linearInterpolation(double v0, double v1, double t) throws SRuntimeException
  {
    // Doit satisfaire la contrainte sur t
    if(t < 0.0 || t > 1.0)
      throw new SRuntimeException("Erreur SMath 002 : Le paramtre d'interpolation t = " + t + " n'est pas compris entre 0 et 1.");
    
    // Calcul de l'interpolation : v = v0*(1 - t) + v1*t
    return v0*(1.0 - t) + (v1*t);
  }
  
  /**
   * ...
   * 
   * @param array
   * @return
   */
  public static double strategicArraySum(double[] array)
  {
    // Mthode #1 : Additionner les termes petits ensemble
    double[] tab_positive = new double[array.length];
    double[] tab_negative = new double[array.length];
    
    int index_positive = 0;
    int index_negative = 0;
    
    // Trier les valeurs positives et ngatives
    for(int i = 0; i < array.length; i++)
    {
      if(array[i] < 0)
      {
        tab_negative[index_negative] = array[i];
        index_negative++;
      }
      else
      {
        tab_positive[index_positive] = array[i];
        index_positive++;
      }
    }
    
    // Ordonner les valeus en ordre croisssant
    Arrays.sort(tab_negative);    // tous les ngatifs gros sont au dbut du tableau
    Arrays.sort(tab_positive);    // tous les zros non affect sont au dbut du tableau
    
    // Faire la somme des nombres ngatifs
    double negative_sum = 0.0;
    
    for(int i = index_negative-1; i >= 0; i--)
      negative_sum += tab_negative[i];
      
    // Faire la somme des nombre positifs
    double positive_sum = 0.0;
    
    for(int i = tab_positive.length - index_positive; i < tab_positive.length; i++)
      positive_sum += tab_positive[i];
    
    // Retourner la somme des positifs avec les ngatifs
    return positive_sum + negative_sum;
  }
  
  /**
   * Mthode dterminant le signe d'un nombre. Les rsultats sont
   * <ul> -1 si a < 0</ul>
   * <ul> 0 si a = 0</ul>
   * <ul> 1 si a > 0</ul>
   * @param a Le nombre.
   * @return Le signe du nombre.
   */
  public static double sgn(double a)
  {
    if(a > 0)
      return 1.0;
    else
      if(a < 0)
        return -1.0;
      else
        return 0.0;
  }
  
  /**
   * Mthode pour obtenir la <b>plus grande valeur</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus grande valeur du tableau.
   */
  public static int findMax(int[] tab)
  {
    int max = tab[0]; 
    
    for(int v : tab)
      if(v > max)
        max = v;
    
    return max;
  }
  
  /**
   * Mthode pour obtenir la <b>plus grande valeur</b> d'un tableau.
   * 
   * @param tab Le tableau.
   * @return La plus grande valeur du tableau.
   */
  public static double findMax(double[] tab)
  {
    double max = tab[0]; 
    
    for(double v : tab)
      if(v > max)
        max = v;
    
    return max;
  }
  
  /**
   * Mthode pour obtenir la <b>plus petite valeur</b> d'un tableau.
   * @param tab Le tableau.
   * @return La plus petite valeur du tableau.
   */
  public static int findMin(int[] tab)
  {
    int min = tab[0]; 
    
    for(int v : tab)
      if(v < min)
        min = v;
    
    return min;
  }
  
  /**
   * Mthode pour obtenir la <b>plus petite valeur</b> d'un tableau.
   * @param tab Le tableau.
   * @return La plus petite valeur du tableau.
   */
  public static double findMin(double[] tab)
  {
    double min = tab[0]; 
    
    for(double v : tab)
      if(v < min)
        min = v;
    
    return min;
  }
  
  /**
   * Mthode pour convertir un tableau d'entier vers un tableau de nombre rel entre une valeur minimale et maximale.
   * La correspondance entre les valeurs converties sera linaire.
   * 
   * @param data Les entiers  convertir.
   * @param min La borne minimale.
   * @param max La borne maximale.
   * @return Le tableau converti.
   * @throws SRuntimeException Si les bornes sont mal dfinies.
   */
  public static double[] mappingIntToDouble(int[] data, double min, double max) throws SRuntimeException
  {
    if(min > max)
      throw new SRuntimeException("Erreur SMath 003 : La borne minimale " + min + " et la borne maximale " + max + " sont mal dfinies.");
    
    int min_value = SMath.findMin(data);
    int max_value = SMath.findMax(data);
    
    double[] result = new double[data.length];
    
    // Itrer sur l'ensemble des lments du tableau.
    for(int i = 0; i < result.length; i++)
    {
      double t = SMath.reverseLinearInterpolation((double)data[i], (double)min_value, (double)max_value);
      result[i] = SMath.linearInterpolation(min, max, t);
    }
    
    return result;
  }
  
  /**
   * Mthode pour convertir un tableau de nombre rel vers un tableau d'entier entre une valeur minimale et maximale.
   * La correspondance entre les valeurs converties sera linaire.
   * 
   * @param data Les donnes  convertir.
   * @param min La borne minimale.
   * @param max La borne maximale.
   * @return Le tableau converti.
   * @throws SRuntimeException Si les bornes sont mal dfinies.
   */
  public static int[] mappingDoubleToInt(double[] data, int min, int max)
  {
    if(min > max)
      throw new SRuntimeException("Erreur SMath 004 : La borne minimale " + min + " et la borne maximale " + max + " sont mal dfinies.");
    
    double min_value = SMath.findMin(data);
    double max_value = SMath.findMax(data);
    
    int[] result = new int[data.length];
    
    // Itrer sur l'ensemble des lments du tableau.
    for(int i = 0; i < result.length; i++)
    {
      double t = SMath.reverseLinearInterpolation(data[i], min_value, max_value);
      result[i] = (int)SMath.linearInterpolation(min, max, t);
    }
    
    return result;
  }
  
}//fin de la classe SMath
