/**
 * 
 */
package sim.physics;

import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.geometry.SGeometricDiscretization;
import sim.math.SVector3d;
import sim.util.SAbstractReadableWriteable;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SParticlesSystem</b> reprsente un systme de particle.
 * 
 * @author Simon Vzina
 * @since 2017-05-31
 * @version 2017-09-20 (labo - La loi de Coulomb v1.1)
 */
public class SParticlesSystem extends SAbstractReadableWriteable {

  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { SKeyWordDecoder.KW_PARTICLE };
  
  /**
   * La variable <b>particlesLi</b> reprsente la liste des particules du systme.
   */
  private final List<SParticle> particles_list;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un systme de particule  partir d'un tableau de particule.
   * 
   * @param particles La tableau des particles.
   */
  public SParticlesSystem(SParticle ... particles)
  {
    // Crer la liste interne des particules du systme. 
    particles_list = new ArrayList<SParticle>();
    
    // Remplir la liste des particules.
    for(SParticle p : particles)
      particles_list.add(p);
  }
  
  /**
   * Constructeur d'un systme de particule  partir d'une liste de particule.
   * 
   * @param list La liste des particules.
   */
  public SParticlesSystem(List<SParticle> list)
  {
    // Crer la liste interne des particules du systme. 
    particles_list = new ArrayList<SParticle>();
    
    // Remplir la liste des particules.
    for(SParticle p : list)
      particles_list.add(p);
  }
  
  
  /**
   * Constructeur d'un systme de particule  partir de la lecture d'un fichier txt dcrivant une lise de particule.
   * 
   * @param file_name Le nom du fichier.
   * @throws SConstructorException S'il y a eu une erreur le de la construction de l'objet.
   */
  public SParticlesSystem(String file_name) throws SConstructorException
  {
    // Crer la liste interne des particules du systme. 
    particles_list = new ArrayList<SParticle>();
    
    try{
      read(file_name);
    }catch(FileNotFoundException e){
      throw new SConstructorException("Erreur SParticlesSystem 001 : Le fichier " + file_name + " n'a pas t trouv.", e);
    }catch(IOException e){
      throw new SConstructorException("Erreur SParticlesSystem 002 : Une erreur de type I/O s'est produite lors de l'analyse du fichier " + file_name + ".", e);
    }
  }
  
  //-----------
  // MTHODE //
  //-----------
  
  /**
   * Mthode pour obtenir le nombre de particules dans le systme.
   * 
   * @return Le nombre de particules dans le systme.
   */
  public int getNbParticles()
  {
    return particles_list.size();
  }
  
  /**
   * Mthode pour ajouter une particule au systme. Cette particule aura ajoute  la fin de la liste des particule du systme (index maximal).
   * 
   * @param p La particule ajoute au systme.
   */
  public void add(SParticle p)
  {
    particles_list.add(p);
  }
  
  /**
   * Mthode pour retirer la dernire particule insere dans le systme.
   * 
   * @return La dernire particule insere dans le systme et qui fut retire.
   * @throws UnsupportedOperationException Si le systme ne possde pas de particule.
   */
  public SParticle removeLast() throws UnsupportedOperationException
  {
    // Vrifier que la liste n'est pas vide.
    if(particles_list.size() == 0)
      throw new UnsupportedOperationException("Erreur SParticlesSystem 003 : La liste de particule est vide.");
      
    return particles_list.remove(particles_list.size()-1);
  }
  
  /**
   * Mthode pour obtenir la force lectrique applique sur une particule par les autres charges du systme.
   * 
   * @param index L'indice de la particule qui va subir la force lectrique.
   * @return La force lectrique applique sur la particule.
   * @throws IndexOutOfBoundsException Si l'indice ne concorde pas avec une particule du systme.
   */
  public SVector3d evaluateForce(int index) throws IndexOutOfBoundsException
  {
    throw new SNoImplementationException("La mthode n'a pas t implmente.");
  }
    
  /**
   * Mthode pour construire un systme de particule  partir d'une liste de positionnment des particules composant le systme.
   * 
   * @param Q La charge totale.
   * @param list La liste des particules.
   * @return Le systme de particules.
   */
  public static SParticlesSystem buildSystem(double Q, List<SVector3d> list)
  {
    // Allocation de la mmoire au systme.
    SParticlesSystem system = new SParticlesSystem();
    
    // Dterminer la charge de chaque particule.
    double Q_particle = Q / (double)list.size();
    
    // Construire les particules.
    for(SVector3d v : list)
      system.add(new SParticle(Q_particle, v));
    
    return system;
  }
  
  /**
   * Mthode pour construire un systme de particules situ le long d'une ligne aligne sur l'axe x positif.
   * 
   * @param Q La charge totale.
   * @param size La taille de la ligne.
   * @param N Le nombre de particules le long de la ligne (la discrtisation).
   * @return Le systme de particles.
   * @throws SRuntimeException S'il y a eu une erreur dans les paramtres.
   */
  public static SParticlesSystem buildLineX(double Q, double size, int N) throws SRuntimeException
  {
    return buildSystem(Q, SGeometricDiscretization.positiveLineXDiscretization(size, N));
  }
  
  /**
   * Mthode pour construire un systme de particules situ sur un carr situ dans le plan XY positif.
   * 
   * @param Q La charge totale.
   * @param size La taille des cts du carr.
   * @param N Le nombre de particules sur le carr dans le plan XY (la discrtisation).
   * @return Le systme de particles.
   * @throws SRuntimeException S'il y a eu une erreur dans les paramtres.
   */
  public static SParticlesSystem buildSquareXY(double Q, double size, int N) throws SRuntimeException
  {
    return buildSystem(Q, SGeometricDiscretization.positiveSquareXYDiscretization(size, (int)Math.sqrt(N)));
  }
  
  @Override
  public void write(BufferedWriter bw) throws IOException 
  {
    bw.write(SKeyWordDecoder.KW_PARTICLES_SYSTEM);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    //crire les proprits de la classe SSphereGeometry et ses paramtres hrits
    writeSParticlesSystemParameter(bw);
    
    bw.write(SKeyWordDecoder.KW_END);
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  /**
   * Mthode pour crire les paramtres associs  la classe SParticlesSystem et ses paramtres hrits.
   * 
   * @param bw Le BufferedWriter crivant l'information dans un fichier txt.
   * @throws IOException Si une erreur I/O s'est produite.
   * @see IOException
   */
  protected void writeSParticlesSystemParameter(BufferedWriter bw) throws IOException
  {
    // Faire l'criture de toutes les particules.
    for(SParticle p : particles_list)
      p.write(bw);
  }
  
  @Override
  public String getReadableName() 
  {
    return SKeyWordDecoder.KW_PARTICLES_SYSTEM;
  }

  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
  @Override
  protected void readingInitialization() throws SInitializationException 
  {
   
  }

  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException, IOException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_PARTICLE : particles_list.add(new SParticle(sbr)); return true;
      
      default : return false;
    }
  }

  @Override
  public String toString() 
  {
    StringBuffer bString = new StringBuffer();
    
    for(int i = 0; i < particles_list.size(); i++)
    {
      bString.append((i+1) + "- ").append(particles_list.get(i).toString()).append("\n");
    }
    
    return bString.toString();
  }

}// fin de la classe SParticlesSystem
